__author__ = 'maartenbreddels'
import os
import sys
import collections
import logging
import concurrent.futures
import threading

import numpy as np
import pyarrow as pa

import vaex
import vaex.utils
import vaex.execution
from vaex.column import ColumnStringArrow, _to_string_sequence


max_length = int(1e5)

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
try:
    import h5py
except:
    if not on_rtd:
        raise
# from vaex.dataset import DatasetLocal

logger = logging.getLogger("vaex.export")
progress_lock = threading.Lock()

class ProgressStatus(object):
    pass

def _export(dataset_input, dataset_output, random_index_column, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True, parallel=True):
    """
    :param DatasetLocal dataset: dataset to export
    :param str path: path for file
    :param lis[str] column_names: list of column names to export or None for all columns
    :param str byteorder: = for native, < for little endian and > for big endian
    :param bool shuffle: export rows in random order
    :param bool selection: export selection or not
    :param progress: progress callback that gets a progress fraction as argument and should return True to continue
    :return:
    """

    if selection:
        if selection == True:  # easier to work with the name
            selection = "default"

    N = len(dataset_input) if not selection else dataset_input.selected_length(selection)
    if N == 0:
        raise ValueError("Cannot export empty table")

    if shuffle and sort:
        raise ValueError("Cannot shuffle and sort at the same time")

    if shuffle:
        shuffle_array = dataset_output.columns[random_index_column]

    partial_shuffle = shuffle and len(dataset_input) != N

    order_array = None
    order_array_inverse = None

    # for strings we also need the inverse order_array, keep track of that
    has_strings = any([dataset_input.is_string(k) for k in column_names])

    if partial_shuffle:
        # if we only export a portion, we need to create the full length random_index array, and
        shuffle_array_full = np.random.choice(len(dataset_input), len(dataset_input), replace=False)
        # then take a section of it
        shuffle_array[:] = shuffle_array_full[shuffle_array_full < N]
        del shuffle_array_full
        order_array = shuffle_array
    elif shuffle:
        # better to do this in memory
        shuffle_array_memory = np.random.choice(N, N, replace=False)
        shuffle_array[:] = shuffle_array_memory
        order_array = shuffle_array
    if order_array is not None:
        indices_r = np.zeros_like(order_array)
        indices_r[order_array] = np.arange(len(order_array))
        order_array_inverse = indices_r
        del indices_r

    if sort:
        if selection:
            raise ValueError("sorting selections not yet supported")
        # these indices sort the input array, but we evaluate the input in sequential order and write it out in sorted order
        # e.g., not b[:] = a[indices]
        # but b[indices_r] = a
        logger.info("sorting...")
        indices = np.argsort(dataset_input.evaluate(sort))
        indices_r = np.zeros_like(indices)
        indices_r[indices] = np.arange(len(indices))
        if has_strings:
            # in this case we already have the inverse ready
            order_array_inverse = indices if ascending else indices[:--1]
        else:
            del indices
        order_array = indices_r if ascending else indices_r[::-1]
        logger.info("sorting done")

    if progress == True:
        progress = vaex.utils.progressbar_callable(title="exporting")
    progress = progress or (lambda value: True)
    progress_total = len(column_names) * len(dataset_input)
    progress_status = ProgressStatus()
    progress_status.cancelled = False
    progress_status.value = 0
    if selection:
        dataset_input.count(selection=selection)  # fill cache for filter and selection
    else:
        len(dataset_input)  # fill filter cache

    sparse_groups = collections.defaultdict(list)
    sparse_matrices = {}  # alternative to a set of matrices, since they are not hashable
    string_columns = []
    futures = []

    thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=1) 
    if True:
        for column_name in column_names:
            sparse_matrix = dataset_output._sparse_matrix(column_name)
            if sparse_matrix is not None:
                # sparse columns are written differently
                sparse_groups[id(sparse_matrix)].append(column_name)
                sparse_matrices[id(sparse_matrix)] = sparse_matrix
                continue
            logger.debug("  exporting column: %s " % column_name)
            future = thread_pool.submit(_export_column, dataset_input, dataset_output, column_name,
                shuffle, sort, selection, N, order_array, order_array_inverse, progress_status, parallel=parallel)
            futures.append(future)

    done = False
    progress(0)
    while not done:
        done = True
        for future in futures:
            try:
                future.result(0.1/4)
            except concurrent.futures.TimeoutError:
                done = False
                break
        if not done:
            if not progress(progress_status.value / float(progress_total)):
                progress_status.cancelled = True
    if not progress_status.cancelled:
        progress(1)

    for sparse_matrix_id, column_names in sparse_groups.items():
        sparse_matrix = sparse_matrices[sparse_matrix_id]
        for column_name in column_names:
            assert not shuffle
            assert selection in [None, False]
            column = dataset_output.columns[column_name]
            column.matrix.data[:] = dataset_input.columns[column_name].matrix.data
            column.matrix.indptr[:] = dataset_input.columns[column_name].matrix.indptr
            column.matrix.indices[:] = dataset_input.columns[column_name].matrix.indices
    return column_names

def _export_column(dataset_input, dataset_output, column_name, shuffle, sort, selection, N,
    order_array, order_array_inverse, progress_status, parallel=True):

        if 1:
            to_array = dataset_output.columns[column_name]
            dtype = dataset_input.data_type(column_name)
            is_string = dtype.is_string
            if is_string:
                # assert isinstance(to_array, pa.Array)  # we don't support chunked arrays here
                # TODO legacy: we still use ColumnStringArrow to write, find a way to do this with arrow
                # this is the case with hdf5 and remote storage
                if not isinstance(to_array, ColumnStringArrow):
                    to_array = ColumnStringArrow.from_arrow(to_array)
            if shuffle or sort:  # we need to create a in memory copy, otherwise we will do random writes which is VERY inefficient
                to_array_disk = to_array
                if np.ma.isMaskedArray(to_array):
                    to_array = np.empty_like(to_array_disk)
                else:
                    if vaex.array_types.is_string_type(dtype):
                        # we create an empty column copy
                        to_array = to_array._zeros_like()
                    else:
                        to_array = np.zeros_like(to_array_disk)
            to_offset = 0  # we need this for selections
            to_offset_unselected = 0 # we need this for filtering
            count = len(dataset_input)# if not selection else dataset_input.length_unfiltered()
            # TODO: if no filter, selection or mask, we can choose the quick path for str
            string_byte_offset = 0

            for i1, i2, values in dataset_input.evaluate(column_name, chunk_size=max_length, filtered=True, parallel=parallel, selection=selection, array_type='numpy-arrow'):
                logger.debug("from %d to %d (total length: %d, output length: %d)", i1, i2, len(dataset_input), N)
                no_values = len(values)
                if no_values:
                    if is_string:
                        # for strings, we don't take sorting/shuffling into account when building the structure
                        to_column = to_array
                        from_sequence = _to_string_sequence(values)
                        to_sequence = to_column.string_sequence.slice(to_offset, to_offset+no_values, string_byte_offset)
                        string_byte_offset += to_sequence.fill_from(from_sequence)
                        to_offset += no_values
                    else:
                        fill_value = np.nan if dtype.kind == "f" else None
                        # assert np.ma.isMaskedArray(to_array) == np.ma.isMaskedArray(values), "to (%s) and from (%s) array are not of both masked or unmasked (%s)" %\
                        # (np.ma.isMaskedArray(to_array), np.ma.isMaskedArray(values), column_name)
                        if shuffle or sort:
                            target_set_item = order_array[i1:i2]
                        else:
                            target_set_item = slice(to_offset, to_offset + no_values)
                        if dtype.is_datetime:
                            values = values.view(np.int64)
                        if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                            to_array.data[target_set_item] = values.filled(fill_value)
                            to_array.mask[target_set_item] = values.mask
                        elif not np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(values):
                            to_array[target_set_item] = values.filled(fill_value)
                        else:
                            to_array[target_set_item] = values
                        to_offset += no_values

                with progress_lock:
                    progress_status.value += i2 - i1
                if progress_status.cancelled:
                    break
                #if not progress(progress_value / float(progress_total)):
                #    break
            if is_string:  # write out the last index
                to_column = to_array
                if selection:
                    to_column.indices[to_offset] = string_byte_offset
                else:
                    to_column.indices[count] = string_byte_offset
            if shuffle or sort:  # write to disk in one go
                if is_string:  # strings are sorted afterwards
                    view = to_array.string_sequence.lazy_index(order_array_inverse)
                    to_array_disk.string_sequence.fill_from(view)
                else:
                    if np.ma.isMaskedArray(to_array) and np.ma.isMaskedArray(to_array_disk):
                        to_array_disk.data[:] = to_array.data
                        to_array_disk.mask[:] = to_array.mask
                    else:
                        to_array_disk[:] = to_array


def export_hdf5_v1(dataset, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True):
    kwargs = locals()
    import vaex.hdf5.export
    vaex.hdf5.export.export_hdf5_v1(**kwargs)


def export_hdf5(dataset, path, column_names=None, byteorder="=", shuffle=False, selection=False, progress=None, virtual=True, sort=None, ascending=True, parallel=True):
    kwargs = locals()
    import vaex.hdf5.export
    vaex.hdf5.export.export_hdf5(**kwargs)


if __name__ == "__main__":
    sys.exit(main(sys.argv))





import pyarrow.fs


import vaex.utils
gcsfs = vaex.utils.optional_import('gcsfs', '>=0.6.2')
import vaex.file.cache
from . import split_options, split_scheme
from .cache import FileSystemHandlerCached

normal_open = open


def glob(path, fs_options={}):
    if '?' in path:
        __, query = path[:path.index('?')], path[path.index('?'):]
    else:
        query = ''
    path, options = split_options(path, fs_options)
    fs = gcsfs.GCSFileSystem(**options)
    return ['gs://' + k + query for k in fs.glob(path)]


def parse(path, fs_options={}, for_arrow=False):
    path = path.replace('fsspec+gs://', 'gs://')
    path, fs_options = split_options(path, fs_options)
    scheme, path = split_scheme(path)
    assert scheme == 'gs'
    use_cache = fs_options.pop('cache', 'true') in [True, 'true', 'True', '1']
    fs = gcsfs.GCSFileSystem(**fs_options)
    fs = pyarrow.fs.FSSpecHandler(fs)
    if use_cache:
        fs = FileSystemHandlerCached(fs, scheme='gs', for_arrow=for_arrow)
    if for_arrow:
        fs = pyarrow.fs.PyFileSystem(fs)
    return fs, path



import vaex
import vaex.array_types
import numpy as np
import pyarrow as pa


class DataType:
    """Wraps numpy and arrow data types in a uniform interface

    Examples:
    >>> import numpy as np
    >>> import pyarrow as pa
    >>> type1 = DataType(np.dtype('f8'))
    >>> type1
    float64
    >>> type2 = DataType(np.dtype('>f8'))
    >>> type2
    float64
    >>> type1 in [float, int]
    True

    """
    def __init__(self, internal):
        self.internal = internal

    def to_native(self):
        '''Removes non-native endianness'''
        return DataType(vaex.utils.to_native_dtype(self.internal))        

    def __eq__(self, other):
        if other == str:
            return self.is_string
        if other == float:
            return self.is_float
        if other == int:
            return self.is_integer
        if other == list:
            return self.is_list
        if isinstance(other, str):
            tester = 'is_' + other
            if hasattr(self, tester):
                return getattr(self, tester)
        if not isinstance(other, DataType):
            other = DataType(other)
        return vaex.array_types.same_type(self.internal, other.internal)

    def __repr__(self):
        '''Standard representation for datatypes


        >>> dtype = DataType(pa.float64())
        >>> dtype.internal
        DataType(double)
        >>> dtype
        float64
        >>> DataType(pa.float32())
        float32
        '''

        internal = self.internal
        if isinstance(internal, np.dtype):
            if internal.byteorder == ">":
                internal = internal.newbyteorder()
        if self.is_datetime:
            internal = self.numpy

        repr = str(internal)
        translate = {
            'datetime64': 'datetime64[ns]',  # for consistency, always add time unit
            'double': 'float64',  # this is what arrow does
            'float': 'float32',  # this is what arrow does
        }
        return translate.get(repr, repr)

    @property
    def name(self):
        '''Alias of dtype.numpy.name'''
        return self.numpy.name

    @property
    def kind(self):
        return self.numpy.kind

    @property
    def numpy(self):
        '''Return the numpy equivalent type

        >>> DataType(pa.float64()).numpy == np.dtype('f8')
        True
        '''
        return vaex.array_types.to_numpy_type(self.internal)

    @property
    def arrow(self):
        '''Return the Apache Arrow equivalent type

        >>> DataType(np.dtype('f8')).arrow == pa.float64()
        True
        '''
        return vaex.array_types.to_arrow_type(self.internal)

    @property
    def is_arrow(self):
        '''Return True if it wraps an Arrow type

        >>> DataType(pa.string()).is_arrow
        True
        >>> DataType(pa.int32()).is_arrow
        True
        >>> DataType(np.dtype('f8')).is_arrow
        False
        '''
        return isinstance(self.internal, pa.DataType)

    @property
    def is_numpy(self):
        '''Return True if it wraps an NumPy dtype

        >>> DataType(np.dtype('f8')).is_numpy
        True
        >>> DataType(pa.string()).is_numpy
        False
        >>> DataType(pa.int32()).is_numpy
        False
        '''
        return isinstance(self.internal, np.dtype)

    @property
    def is_numeric(self):
        '''Tests if type is numerical (float, int)

        >>> DataType(np.dtype('f8')).is_numeric
        True
        >>> DataType(pa.float32()).is_numeric
        True
        '''
        return self.kind in 'fiu'

    @property
    def is_primitive(self):
        '''Tests if type is numerical (float, int, bool)

        >>> DataType(np.dtype('b')).is_primitive
        True
        >>> DataType(pa.bool_()).is_primitive
        True
        '''
        return self.kind in 'fiub'

    @property
    def is_datetime(self):
        """Tests if dtype is datetime (numpy) or timestamp (arrow)

        Date/Time:
        >>> date_type = DataType(np.dtype('datetime64'))
        >>> date_type
        datetime64[ns]
        >>> date_type == 'datetime'
        True

        Using Arrow:

        >>> date_type = DataType(pa.timestamp('ns'))
        >>> date_type
        datetime64[ns]
        >>> date_type == 'datetime'
        True
        """
        if self.is_string:
            return False
        return vaex.array_types.to_numpy_type(self.internal).kind in 'M'

    @property
    def is_timedelta(self):
        '''Test if timedelta

        >>> dtype = DataType(np.dtype('timedelta64'))
        >>> dtype
        timedelta64
        >>> dtype == 'timedelta'
        True
        >>> dtype.is_timedelta
        True
        '''
        return vaex.array_types.to_numpy_type(self.internal).kind in 'm'

    @property
    def is_float(self):
        '''Test if a float (float32 or float64)

        >>> dtype = DataType(np.dtype('float32'))
        >>> dtype
        float32
        >>> dtype == 'float'
        True
        >>> dtype == float
        True
        >>> dtype.is_float
        True

        Using Arrow:
        >>> DataType(pa.float32()) == float
        True
        '''
        return vaex.array_types.to_numpy_type(self.internal).kind in 'f'

    @property
    def is_unsigned(self):
        '''Test if an (unsigned) integer

        >>> dtype = DataType(np.dtype('uint32'))
        >>> dtype
        uint32
        >>> dtype == 'unsigned'
        True
        >>> dtype.is_unsigned
        True

        Using Arrow:
        >>> DataType(pa.uint32()).is_unsigned
        True
        '''
        return vaex.array_types.to_numpy_type(self.internal).kind in 'u'

    @property
    def is_signed(self):
        '''Test if a (signed) integer

        >>> dtype = DataType(np.dtype('int32'))
        >>> dtype
        int32
        >>> dtype == 'signed'
        True

        Using Arrow:
        >>> DataType(pa.int32()).is_signed
        True
        '''
        return vaex.array_types.to_numpy_type(self.internal).kind in 'i'

    @property
    def is_integer(self):
        '''Test if an (unsigned or signed) integer

        >>> DataType(np.dtype('uint32')) == 'integer'
        True
        >>> DataType(np.dtype('int8')) == int
        True
        >>> DataType(np.dtype('int16')).is_integer
        True

        Using Arrow:
        >>> DataType(pa.uint32()).is_integer
        True
        >>> DataType(pa.int16()) == int
        True

        '''
        return vaex.array_types.to_numpy_type(self.internal).kind in 'iu'

    @property
    def is_string(self):
        '''Test if an (arrow) string or large_string

        >>> DataType(pa.string()) == str
        True
        >>> DataType(pa.large_string()) == str
        True
        >>> DataType(pa.large_string()).is_string
        True
        >>> DataType(pa.large_string()) == 'string'
        True
        '''
        return vaex.array_types.is_string_type(self.internal)

    @property
    def is_list(self):
        '''Test if an (arrow) list or large_string

        >>> DataType(pa.list_(pa.string())) == list
        True
        >>> DataType(pa.large_list(pa.string())) == list
        True
        >>> DataType(pa.list_(pa.string())).is_list
        True
        >>> DataType(pa.list_(pa.string())) == 'list'
        True
        '''
        return self.is_arrow and (pa.types.is_list(self.internal) or pa.types.is_large_list(self.internal))

    @property
    def value_type(self):
        '''Return the DataType of the list values

        >>> DataType(pa.list_(pa.string())).value_type
        string
        >>> DataType(pa.list_(pa.float64())).value_type
        float64
        '''
        if not self.is_list:
            raise TypeError(f'{self} is not a list type')
        return DataType(self.internal.value_type)

    def upcast(self):
        '''Cast to the higest data type matching the type

        >>> DataType(np.dtype('uint32')).upcast()
        uint64
        >>> DataType(np.dtype('int8')).upcast()
        int64
        >>> DataType(np.dtype('float32')).upcast()
        float64

        Using Arrow
        >>> DataType(pa.float32()).upcast()
        float64
        '''
        return DataType(vaex.array_types.upcast(self.internal))

    @property
    def byteorder(self):
        return self.numpy.byteorder


import vaex as vx
import sys
import timeit
import numpy as np
import argparse

subspace, dataset, limits = None, None, None


def main(argv):
    global subspace, dataset, limits
    parser = argparse.ArgumentParser(argv[0])
    parser.add_argument("-N", help="run each batch N times (default: %(default)s)", type=int, default=5)
    parser.add_argument("-f", "--fraction", help="fraction of dataset to use (default: %(default)s)", default=1., type=float)
    parser.add_argument("-b", "--buffer-size", default=vx.execution.buffer_size_default, help="buffer size per thread (default: %(default)s)", type=int)
    parser.add_argument("-r", "--repeat", default=5, help="repeat benchmark X times (default: %(default)s)", type=int)
    parser.add_argument("-c", "--cpu-count", default=vx.multithreading.thread_count_default, help="thread count/cpu count (default: %(default)s)", type=int)
    parser.add_argument("filename", help="filename of dataset to use")
    parser.add_argument("expressions", help="list of expressions to export (or all when empty)", nargs="*")

    args = parser.parse_args(argv[1:])
    vx.execution.buffer_size_default = args.buffer_size
    vx.multithreading.thread_count_default = args.cpu_count

    progressbar = False
    fn = args.filename
    # print(("opening", fn))
    dataset = vx.open(fn)
    dataset.set_active_fraction(args.fraction)
    # dataset = vx.open_many(fn)

    expressions = args.expressions
    # print "subspace", expressions
    subspace = dataset(*expressions)
    itemsize = dataset.columns[expressions[0]].dtype.itemsize
    byte_size = len(dataset) * len(expressions) * itemsize
    byte_size1 = len(dataset) * itemsize
    # sums = subspace.sum()

    limits = subspace.minmax()
    print(limits)

    N = args.N
    # print("benchmarking minmax")
    # expr = "subspace.minmax()"
    # expr = "dataset.count('{}')".format(expressions[0])
    # times = timeit.repeat(expr, setup="from vaex.benchmark import subspace, dataset, np", repeat=args.repeat, number=N)
    # print("minimum time", min(times) / N)
    # bandwidth = [byte_size1 / 1024.**3 / (time / N) for time in times]
    # print("%f GiB/s" % max(bandwidth))

    # speed = [len(dataset) / (time / N) / 1e9 for time in times]
    # print("%f billion rows/s " % max(speed))

    print()
    print("benchmarking histogram")
    expr = "dataset.count(binby=['{}', '{}'], limits=limits)".format(*expressions)
    times = timeit.repeat(expr, setup="from vaex.benchmark import subspace, dataset, np, limits", repeat=args.repeat, number=N)
    print("minimum time", min(times) / N)
    bandwidth = [byte_size / 1024.**3 / (time / N) for time in times]
    print("%f GiB/s" % max(bandwidth))

    speed = [len(dataset) / (time / N) / 1e9 for time in times]
    print("%f billion rows/s " % max(speed))
    print()

    # expr = "np.nansum(dataset.columns['x'])"
    # print("sum=", subspace.sum())
    # print("benchmarking sum")
    # expr = "subspace.sum()"
    # times = timeit.repeat(expr, setup="from vaex.benchmark import subspace, dataset, np", repeat=args.repeat, number=N)
    # print("minimum time", min(times) / N)
    # bandwidth = [byte_size / 1024.**3 / (time / N) for time in times]
    # print("%f GiB/s" % max(bandwidth))



import os
import logging
import vaex.utils
import collections

try:
    collections_abc = collections.abc
except AttributeError:
    collections_abc = collections

logger = logging.getLogger("vaex.settings")


class Files(object):
    def __init__(self, open, recent):
        self.open = open
        self.recent = recent


class Settings(object):
    def __init__(self, filename):
        self.filename = filename
        if not os.path.exists(filename):
            with open(filename, "w"):
                pass
        with open(self.filename) as f:
            self.settings = vaex.utils.yaml_load(f)  # yaml.load(f, Loader=Loader)
        if self.settings is None:
            self.settings = {}
        # logger.debug("settings: %r", self.settings)

    def auto_store_dict(self, key):
        # TODO: no nested keys supported yet
        if key not in self.settings:
            self.settings[key] = {}
        return AutoStoreDict(self, self.settings[key])

    def store(self, key, value):
        parts = key.split(".")
        obj = self.settings
        for part in parts[:-1]:
            # print part,
            if part not in obj:
                obj[part] = {}
            obj = obj[part]
        obj[parts[-1]] = value
        # print self.settings
        self.dump()

    def dump(self):
        with open(self.filename, "w") as f:
            # yaml.dump(self.settings, f)
            vaex.utils.yaml_dump(f, self.settings)

    def get(self, key, default=None):
        logger.debug("get %r", key)
        parts = key.split(".")
        obj = self.settings
        for part in parts:
            if part not in obj:
                logger.debug("return %r (default)", default)
                return default
            obj = obj[part]
        logger.debug("return %r", obj)
        return obj


settings = {}


class AutoStoreDict(collections_abc.MutableMapping):
    def __init__(self, settings, store):
        self.store = store
        self.settings = settings

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value
        self.settings.dump()

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]
        self.settings.dump()

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return key


main = Settings(os.path.join(vaex.utils.get_private_dir(), "main.yml"))
webclient = Settings(os.path.join(vaex.utils.get_private_dir(), "webclient.yml"))
webserver = Settings(os.path.join(vaex.utils.get_private_dir(), "webserver.yml"))
cluster = Settings(os.path.join(vaex.utils.get_private_dir(), "cluster.yml"))

# yaml.load()

if __name__ == "__main__":
    webclient.store("bla.la.l", 1)



__author__ = 'breddels'

import unittest
import vaex as vx
import vaex.utils
import vaex.image
import numpy as np
default_size = 2
default_shape = (default_size, default_size)

class TestImage(unittest.TestCase):
    def test_blend(self):
        black = vaex.image.background(default_shape, "black")
        white = vaex.image.background(default_shape, "white")
        transparant_black = black * 1
        transparant_black[...,3] = 100
        for mode in vaex.image.modes:
            self.assert_(np.all(vaex.image.blend([black, white]) == white))
            self.assert_(np.all(vaex.image.blend([white, black]) == black))

            grey = vaex.image.blend([white, transparant_black], mode)

            self.assert_(np.all(grey[...,0:3] < white[...,0:3]))
            self.assert_(np.all(grey[...,0:3] > black[...,0:3]))

    def test_pil(self):
        white = vaex.image.background(default_shape, "white")
        # not much to test
        im = vaex.image.rgba_2_pil(white)
        raw_data = vaex.image.pil_2_data(im)
        url = vaex.image.rgba_to_url(white)


class TestAttrDict(unittest.TestCase):
    def test_attrdict(self):
        d = vx.utils.AttrDict()
        d.a = 1
        self.assertEqual(d.a, 1)
        with self.assertRaises(KeyError):
            a = d.doesnotexist


        d = vx.utils.AttrDict(a=1)
        print(d.__dict__)
        self.assertEqual(d.a, 1)
        with self.assertRaises(KeyError):
            a = d.doesnotexist

if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-
# from sampy import *
# from SocketServer import ThreadingMixIn
# try:
# import sampy
# except ImportError:
# import astropy.samp as sampy
import logging
import threading
import time
import astropy.samp
import getpass
import requests
from requests.auth import HTTPBasicAuth
import astropy.io.votable

logger = logging.getLogger("vaex.samp")


class Samp(object):
    def __init__(self, daemon=True, name=None):
        self.client = astropy.samp.SAMPIntegratedClient(metadata={"samp.name": "Gavi client" if name is None else name,
                                                                  "samp.description.text": "Gavi client" if name is None else name,
                                                                  "gavi.samp.version": "0.01"}, callable=True)

        # sampy doesn't make this thread Daeamon, so the python process never stops on the cmd line
        # this fixes that
        # def _myrun_client():
        # if self.client.client._callable:
        # self.client.client._thread = threading.Thread(target = self.client.client._serve_forever)
        # self.client.client._thread.setDaemon(True)
        # self.client.client._thread.start()
        # if daemon:
        # self.client.client._run_client = _myrun_client
        connected = False
        try:
            self.client.connect()
            connected = True
        except astropy.samp.SAMPHubError as e:
            # print "error connecting to hub", e
            pass

        if connected:
            # self.client.client._thread.setDaemon(False)
            logger.info("connected to SAMP hub")
            logger.info("binding events")
            self.client.bind_receive_call("table.load.votable", self._onTableLoadVotable)
            self.client.bind_receive_notification("table.load.votable", self._onTableLoadVotable)
            # self.client.bindReceiveNotification	("table.highlight.row", self._onSampNotification)
            # self.client.bindReceiveMessage("table.load.votable", self._onSampCall)
            # self.client.bindReceiveResponse("table.load.votable", self._onSampCall)

            # self.client.bindReceiveCall("samp.*", self._onSampCall)
            # self.client.bindReceiveNotification("samp.*", self._onSampNotification)

            # self.client.bindReceiveCall("table.*", self._onSampCall)
            # self.client.bindReceiveNotification("table.*", self._onSampNotification)
            # self.client.bindReceiveMessage("table.*", self._onSampCall)
            # self.client.bindReceiveResponse("table.*", self._onSampCall)

            # self.client.bindReceiveMessage("table.votable.*", self._onSampCall)
            # self.client.bindReceiveResponse("table.votable.*", self._onSampCall)

    # def connect(self):
    # self.client.connect()
        self.tableLoadCallbacks = []

    def _onTableLoadVotable(self, private_key, sender_id, msg_id, mtype, params, extra):
        print(("Msg:", repr(private_key), repr(sender_id), repr(msg_id), repr(mtype), repr(params), repr(extra)))
        try:
            url = params["url"]
            table_id = params["table-id"]
            name = params["name"]
            for callback in self.tableLoadCallbacks:
                callback(url, table_id, name)
        except:
            logger.exception("event handler failed")

        if msg_id is not None:  # if SAMP call, send a reply
            self.client.ereply(msg_id, sampy.SAMP_STATUS_OK, result={"txt": "loaded"})

    def _onSampNotification(self, private_key, sender_id, mtype, params, extra):
        print(("Notification:", repr(private_key), repr(sender_id), repr(mtype), repr(params), repr(extra)))

    def _onSampCall(self, private_key, sender_id, msg_id, mtype, params, extra):
        print("----")
        try:
            print(("Call:", repr(private_key), repr(sender_id), repr(msg_id), repr(mtype), repr(params), repr(extra)))
            self.client.ereply(msg_id, sampy.SAMP_STATUS_OK, result={"txt": "printed"})
        except:
            print("errrrrrrororrrr hans!")


# similar to http://astrofrog-debug.readthedocs.io/en/latest/vo/samp/example_table_image.html
class SampSingle(object):
    def __init__(self, name="vaex - single table load"):
        self.done = False
        self.client = astropy.samp.SAMPIntegratedClient(name=name)
        self.client.connect()

        def call(private_key, sender_id, msg_id, mtype, params, extra):
            self.params = params
            self.client.reply(msg_id, {"samp.status": "samp.ok", "samp.result": {}})
            self.done = True

        def notify(private_key, sender_id, mtype, params, extra):
            self.params = params
            self.done = True

        self.client.bind_receive_call("table.load.votable", call)
        self.client.bind_receive_notification("table.load.votable", notify)

    def wait_for_table(self):
        logging.debug("waiting for samp msg: table.load.votable")
        while not self.done:
            time.sleep(0.1)
        logging.debug("got samp msg: table.load.votable with params: %s", self.params)
        self.client.disconnect()
        return self.params["url"]


try:
    input = raw_input
except:
    pass
# py2/3 comp
try:
    from StringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO


def ask_cmd_line(username, password):
    if not username:
        username = input("Username:")
    if not password:
        password = getpass.getpass("Password:")
    return username, password


def fetch_votable(url, username=None, password=None, ask=ask_cmd_line):
    done = False
    auth = None
    while not done:
        if username and password:
            auth = HTTPBasicAuth(username, password)
        req = requests.get(url, auth=auth)
        if req.status_code == 401:
            result = ask(username, password)
            if result:
                username, password = result
                auth = HTTPBasicAuth(username, password)
            else:
                return
        else:
            done = True
    if req.status_code == 200:
        f = BytesIO()
        f.write(req.content)
        return astropy.io.votable.parse_single_table(f)
    else:
        return req


def single_table(username=None, password=None):
    # from vaex.samp import SampSingle
    # from astropy.table import Table
    samp = SampSingle()
    url = samp.wait_for_table()
    import os
    import tempfile
    import requests

    done = False
    table = fetch_votable(url, username, password)
    return table
    # while not done:
    # r = requests.get()
    # tempdir = tempfile.mkdtemp()
    # path = os.path.join(tempdir, "out.vot.gz")
    # os.system("wget -c -O %s %s" % (path, url))
    # parts = os.path.split(url)
    # print("parts", parts)
    # path2 = os.path.join(tempdir, "out.vot")
    # os.system("gunzip %s > /dev/null; mv %s %s  > /dev/null" % (path, path, path2))
    # print(path, path2)
    # table = Table.read(path2, format="votable")
    # return from_astropy_table(table)



__author__ = 'maartenbreddels'
import os
import mmap
import itertools
import functools
import collections
import logging
import numpy as np
import numpy.ma
import vaex
import astropy.table
import astropy.units
from vaex.utils import ensure_string
import re
import six

from vaex.dataset import DatasetArrays, DatasetFile
logger = logging.getLogger("vaex.file")
import vaex.dataset
import vaex.file
dataset_type_map = {}

from vaex.expression import Expression

from vaex.dataset_mmap import DatasetMemoryMapped


import struct
class HansMemoryMapped(DatasetMemoryMapped):
	def __init__(self, filename, filename_extra=None):
		super(HansMemoryMapped, self).__init__(filename)
		self.pageSize, \
		self.formatSize, \
		self.numberParticles, \
		self.numberTimes, \
		self.numberParameters, \
		self.numberCompute, \
		self.dataOffset, \
		self.dataHeaderSize = struct.unpack("Q"*8, self.mapping[:8*8])
		zerooffset = offset = self.dataOffset
		length = self.numberParticles+1
		stride = self.formatSize//8 # stride in units of the size of the element (float64)

		# TODO: ask Hans for the self.numberTimes-2
		lastoffset = offset + (self.numberParticles+1)*(self.numberTimes-2)*self.formatSize
		t_index = 3
		names = "x y z vx vy vz".split()
		midoffset = offset + (self.numberParticles+1)*self.formatSize*t_index
		names = "x y z vx vy vz".split()

		for i, name in enumerate(names):
			self.addColumn(name+"_0", offset+8*i, length, dtype=np.float64, stride=stride)

		for i, name in enumerate(names):
			self.addColumn(name+"_last", lastoffset+8*i, length, dtype=np.float64, stride=stride)


		names = "x y z vx vy vz".split()

		if 1:
			stride = self.formatSize//8
			#stride1 = self.numberTimes #*self.formatSize/8
			for i, name in enumerate(names):
				# TODO: ask Hans for the self.numberTimes-1
				self.addRank1(name, offset+8*i, length=self.numberParticles+1, length1=self.numberTimes-1, dtype=np.float64, stride=stride, stride1=1)

		if filename_extra is None:
			basename = os.path.basename(filename)
			if os.path.exists(basename + ".omega2"):
				filename_extra = basename + ".omega2"

		if filename_extra is not None:
			self.addFile(filename_extra)
			mapping = self.mapping_map[filename_extra]
			names = "J_r J_theta J_phi Theta_r Theta_theta Theta_phi Omega_r Omega_theta Omega_phi r_apo r_peri".split()
			offset = 0
			stride = 11
			#import pdb
			#pdb.set_trace()
			for i, name in enumerate(names):
				# TODO: ask Hans for the self.numberTimes-1
				self.addRank1(name, offset+8*i, length=self.numberParticles+1, length1=self.numberTimes-1, dtype=np.float64, stride=stride, stride1=1, filename=filename_extra)

				self.addColumn(name+"_0", offset+8*i, length, dtype=np.float64, stride=stride, filename=filename_extra)
				self.addColumn(name+"_last", offset+8*i + (self.numberParticles+1)*(self.numberTimes-2)*11*8, length, dtype=np.float64, stride=stride, filename=filename_extra)


	@classmethod
	def can_open(cls, path, *args, **kwargs):
		return os.path.splitext(path)[-1] == ".bin"
		basename, ext = os.path.splitext(path)
		#if os.path.exists(basename + ".omega2"):
		#	return True
		#return True

	@classmethod
	def get_options(cls, path):
		return []

	@classmethod
	def option_to_args(cls, option):
		return []
dataset_type_map["buist"] = HansMemoryMapped

def _python_save_name(name, used=[]):
	first, rest = name[0], name[1:]
	name = re.sub("[^a-zA-Z_]", "_", first) +  re.sub("[^a-zA-Z_0-9]", "_", rest)
	if name in used:
		nr = 1
		while name + ("_%d" % nr) in used:
			nr += 1
		name = name + ("_%d" % nr)
	return name

def _try_unit(unit):
	try:
		unit = astropy.units.Unit(str(unit))
		if not isinstance(unit, astropy.units.UnrecognizedUnit):
			return unit
	except:
		#logger.exception("could not parse unit: %r", unit)
		pass
	try:
		unit_mangle = re.match(".*\[(.*)\]", str(unit)).groups()[0]
		unit = astropy.units.Unit(unit_mangle)
	except:
		pass#logger.exception("could not parse unit: %r", unit)
	if isinstance(unit, six.string_types):
		return None
	elif isinstance(unit, astropy.units.UnrecognizedUnit):
		return None
	else:
		return unit


class SoneiraPeebles(DatasetArrays):
	def __init__(self, dimension, eta, max_level, L):
		columns = {}
		#InMemory.__init__(self)
		def todim(value):
			if isinstance(value, (tuple, list)):
				assert len(value) >= dimension, "either a scalar or sequence of length equal to or larger than the dimension"
				return value[:dimension]
			else:
				return [value] * dimension

		eta = eta
		max_level = max_level
		N = eta**(max_level)
		# array[-1] is used as a temp storage
		array = np.zeros((dimension+1, N), dtype=np.float64)
		L = todim(L)

		for d in range(dimension):
			vaex.vaexfast.soneira_peebles(array[d], 0, 1, L[d], eta, max_level)
		for d, name in zip(list(range(dimension)), "x y z w v u".split()):
			columns[name] = array[d]
		super(SoneiraPeebles, self).__init__(columns)

# dataset_type_map["soneira-peebles"] = SoneiraPeebles


class Zeldovich(DatasetArrays):
	def __init__(self, dim=2, N=256, n=-2.5, t=None, seed=None, scale=1, name="zeldovich approximation"):
		super(Zeldovich, self).__init__(name=name)

		if seed is not None:
			np.random.seed(seed)
		#sys.exit(0)
		shape = (N,) * dim
		A = np.random.normal(0.0, 1.0, shape)
		F = np.fft.fftn(A)
		K = np.fft.fftfreq(N, 1./(2*np.pi))[np.indices(shape)]
		k = (K**2).sum(axis=0)
		k_max = np.pi
		F *= np.where(np.sqrt(k) > k_max, 0, np.sqrt(k**n) * np.exp(-k*4.0))
		F.flat[0] = 0
		#pylab.imshow(np.where(sqrt(k) > k_max, 0, np.sqrt(k**-2)), interpolation='nearest')
		grf = np.fft.ifftn(F).real
		Q = np.indices(shape) / float(N-1) - 0.5
		s = np.array(np.gradient(grf)) / float(N)
		#pylab.imshow(s[1], interpolation='nearest')
		#pylab.show()
		s /= s.max() * 100.
		#X = np.zeros((4, 3, N, N, N))
		#for i in range(4):
		#if t is None:
		#	s = s/s.max()
		t = t or 1.
		X = Q + s * t

		for d, name in zip(list(range(dim)), "xyzw"):
			self.add_column(name, X[d].reshape(-1) * scale)
		for d, name in zip(list(range(dim)), "xyzw"):
			self.add_column("v"+name, s[d].reshape(-1) * scale)
		for d, name in zip(list(range(dim)), "xyzw"):
			self.add_column(name+"0", Q[d].reshape(-1) * scale)
		return

# dataset_type_map["zeldovich"] = Zeldovich


class AsciiTable(DatasetMemoryMapped):
	def __init__(self, filename):
		super(AsciiTable, self).__init__(filename, nommap=True)
		import asciitable
		table = asciitable.read(filename)
		logger.debug("done parsing ascii table")
		#import pdb
		#pdb.set_trace()
		#names = table.array.dtype.names
		names = table.dtype.names

		#data = table.array.data
		for i in range(len(table.dtype)):
			name = table.dtype.names[i]
			type = table.dtype[i]
			if type.kind in ["f", "i"]: # only store float and int
				#datagroup.create_dataset(name, data=table.array[name].astype(np.float64))
				#dataset.addMemoryColumn(name, table.array[name].astype(np.float64))
				self.addColumn(name, array=table[name])
		#dataset.samp_id = table_id
		#self.list.addDataset(dataset)
		#return dataset

	@classmethod
	def can_open(cls, path, *args, **kwargs):
		can_open = path.endswith(".asc")
		logger.debug("%r can open: %r"  %(cls.__name__, can_open))
		return can_open






__author__ = 'maartenbreddels'
import sys
import vaex.utils
import vaex as vx
import os
# data_dir = "/tmp/vaex/data"
data_dir = vaex.utils.get_private_dir("data")

try:
    from urllib import urlretrieve  # py2
except ImportError:
    from urllib.request import urlretrieve  # py3


def _url_to_filename(url, replace_ext=None, subdir=None):
    if subdir:
        filename = os.path.join(data_dir, subdir, url.split("/")[-1])
    else:
        filename = os.path.join(data_dir, url.split("/")[-1])
    if replace_ext:
        dot_index = filename.rfind(".")
        filename = filename[:dot_index] + replace_ext
    return filename


class Hdf5Download(object):
    def __init__(self, url):
        self.url = url
        self.url_list = [self.url]

    @property
    def filename(self):
        return os.path.join(data_dir, _url_to_filename(self.url))

    def download(self, force=False):
        if not os.path.exists(self.filename) or force:
            print("Downloading %s to %s" % (self.url, self.filename))
            code = os.system(self.wget_command(0))
            if not os.path.exists(self.filename):
                print("wget failed, using urlretrieve")
                self.download_urlretrieve()

    def download_urlretrieve(self):
        urlretrieve(self.url, self.filename)

    def fetch(self, force_download=False):
        self.download(force=force_download)
        return vx.open(self.filename)

    def wget_command(self, i):
        assert i == 0
        url = self.url_list[i]
        return "wget --progress=bar:force -c -P %s %s" % (data_dir, url)


class NYCTaxi(object):
    def __init__(self, name, url_list, subdir="nyc_taxi"):
        self.name = name
        self.subdir = subdir
        self.url_list = url_list

    @property
    def filename_single(self):
        return os.path.join(data_dir, self.subdir, self.name + ".hdf5")

    @property
    def filenames(self):
        return list(map(lambda url: _url_to_filename(url, subdir=self.subdir), self.url_list))

    @property
    def filenames_vaex(self):
        return list(map(lambda x: _url_to_filename(x, ".hdf5", subdir=self.subdir), self.url_list))

    def fetch(self):
        ds = self.fetch_multi()
        if len(self.filenames) > 1:
            if not os.path.exists(self.filename_single):
                ds.export_hdf5(self.filename_single)
            ds = vx.open(self.filename_single)
        return ds

    def fetch_multi(self):
        self.download()
        self.fix()
        self.convert()
        return self.open()

    def wget_command(self, i):
        url = self.url_list[i]
        return "wget -c -P %s %s" % (os.path.join(data_dir, self.subdir), url)

    def download_urlretrieve(self, i):
        urlretrieve(self.url_list[i], self.filenames[i])

    def download(self, force=False):
        for i, url in enumerate(self.url_list):
            if not os.path.exists(self.filenames[i]) or force:
                print("Downloading %s (%d out of %d)" % (url, i + 1, len(self.url_list)))
                # code = os.system("wget -c -P %s %s" % (os.path.join(data_dir, self.subdir), url))
                code = os.system(self.wget_command(i))
                if not os.path.exists(self.filenames[i]):
                    print("wget failed, try urlretrieve")
                    self.download_urlretrieve(i)
                    if not os.path.exists(self.filenames[i]):
                        raise RuntimeError("could not download file: " + self.filenames[i])

    def fix(self):
        pass

    def convert(self, force=False):
        import pandas as pd
        skips = ["store_and_fwd_flag"]
        for i, (input, output) in enumerate(zip(self.filenames, self.filenames_vaex)):
            date_names = ["tpep_pickup_datetime", "tpep_dropoff_datetime"]
            if not os.path.exists(output) or force:
                print("Converting %s to %s (%d out of %d)" % (input, output, i + 1, len(self.filenames)))
                df = pd.read_csv(input, parse_dates=date_names)

                for skip in skips:
                    if skip in df:
                        del df["store_and_fwd_flag"]
                ds = vx.from_pandas(df)
                ds.add_virtual_column("pickup_hour", "dt_hour(tpep_pickup_datetime)")
                ds.add_virtual_column("dropoff_hour", "dt_hour(tpep_dropoff_datetime)")
                ds.add_virtual_column("pickup_dayofweek", "dt_dayofweek(tpep_pickup_datetime)")
                ds.add_virtual_column("dropoff_dayofweek", "dt_dayofweek(tpep_dropoff_datetime)")
                ds.select("(pickup_longitude != 0) & (pickup_latitude != 0) & (dropoff_longitude != 0) & (dropoff_latitude != 0)")
                ds.export_hdf5(output, virtual=True, selection=True)

    def open(self):
        return vx.open_many(self.filenames_vaex) if len(self.filenames_vaex) != 1 else vx.open(self.filenames_vaex[0])


urllist = """https://storage.googleapis.com/tlc-trip-data/2013/green_tripdata_2013-08.csv
https://storage.googleapis.com/tlc-trip-data/2013/green_tripdata_2013-09.csv
https://storage.googleapis.com/tlc-trip-data/2013/green_tripdata_2013-10.csv
https://storage.googleapis.com/tlc-trip-data/2013/green_tripdata_2013-11.csv
https://storage.googleapis.com/tlc-trip-data/2013/green_tripdata_2013-12.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-01.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-02.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-03.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-04.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-05.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-06.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-07.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-08.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-09.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-10.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-11.csv
https://storage.googleapis.com/tlc-trip-data/2014/green_tripdata_2014-12.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-01.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-02.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-03.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-04.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-05.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-06.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-07.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-08.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-09.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-10.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-11.csv
https://storage.googleapis.com/tlc-trip-data/2015/green_tripdata_2015-12.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-01.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-02.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-03.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-04.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-05.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-06.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-07.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-08.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-09.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-10.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-11.csv
https://storage.googleapis.com/tlc-trip-data/2009/yellow_tripdata_2009-12.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-01.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-02.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-03.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-04.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-05.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-06.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-07.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-08.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-09.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-10.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-11.csv
https://storage.googleapis.com/tlc-trip-data/2010/yellow_tripdata_2010-12.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-01.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-02.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-03.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-04.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-05.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-06.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-07.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-08.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-09.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-10.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-11.csv
https://storage.googleapis.com/tlc-trip-data/2011/yellow_tripdata_2011-12.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-01.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-02.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-03.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-04.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-05.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-06.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-07.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-08.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-09.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-10.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-11.csv
https://storage.googleapis.com/tlc-trip-data/2012/yellow_tripdata_2012-12.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-01.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-02.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-03.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-04.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-05.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-06.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-07.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-08.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-09.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-10.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-11.csv
https://storage.googleapis.com/tlc-trip-data/2013/yellow_tripdata_2013-12.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-01.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-02.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-03.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-04.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-05.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-06.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-07.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-08.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-09.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-10.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-11.csv
https://storage.googleapis.com/tlc-trip-data/2014/yellow_tripdata_2014-12.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-01.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-02.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-03.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-04.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-05.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-06.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-07.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-08.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-09.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-10.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-11.csv
https://storage.googleapis.com/tlc-trip-data/2015/yellow_tripdata_2015-12.csv""".split("\n")

nyctaxi_yellow_2015_jan = NYCTaxi("nyc_taxi2015jan", [urllist[-12]])
nyctaxi_yellow_2015 = NYCTaxi("nyc_taxi2015", urllist[-12:])
nyctaxi_yellow_201x = NYCTaxi("nyc_taxi_all", [url for url in urllist if "yellow" in url and "201" in url])

helmi_de_zeeuw = Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/helmi-dezeeuw-2000-FeH-v2.hdf5")
helmi_de_zeeuw_10percent = Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/helmi-dezeeuw-2000-FeH-v2-10percent.hdf5")
tgas = Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/tgas.hdf5")
tgas_1percent = Hdf5Download("https://github.com/vaexio/vaex-datasets/releases/download/v1.0/tgas_1percent.hdf5")



import pyarrow as pa

import vaex.arrow.convert
import vaex.multithreading
import vaex.utils


process_count_default = vaex.utils.get_env_type(int, 'VAEX_NUM_PROCESSES', vaex.multithreading.thread_count_default)


_pool = None
def _get_pool():
    global _pool
    global _mempool
    if _pool is None:
        from multiprocessing import Pool
        _pool = Pool(process_count_default)
    return _pool


def _trim(ar):
    if isinstance(ar, vaex.array_types.supported_arrow_array_types):
        ar = vaex.arrow.convert.trim_buffers_for_pickle(ar)
        pass
    return ar


def apply(f, args, kwargs, multiprocessing):
    if multiprocessing:
        args = [_trim(k) for k in args]
        kwargs = {k:_trim(v) for k, v in kwargs.items()}
        result = _get_pool().apply(f, args, kwargs)
        return result
    else:
        return f(*args, **kwargs)



import functools
import numpy
import dask.base
import logging
import uuid


log = logging.getLogger('vaex.cache')


import vaex.utils

def fingerprint(*args, **kwargs):
    try:
        _restore = uuid.uuid4
        def explain(*args):
            raise TypeError('You have passed in an object for which we cannot determine a fingerprint')
        uuid.uuid4 = explain
        return dask.base.tokenize(*args, **kwargs)
    finally:
        uuid.uuid4 = _restore


def output_file(callable=None, path_input=None, fs_options_input={}, fs_input=None, path_output=None, fs_options_output={}, fs_output=None):
    """Decorator to do cached conversion from path_input → path_output.

    Caching is active if with the input file, the *args, and **kwargs are unchanged, otherwise callable
    will be called again.


    """
    def wrapper1(callable=callable):
        def wrapper2(callable=callable):
            def call(callable=callable, *args, **kwargs):
                base, ext, fs_options_output_ = vaex.file.split_ext(path_output, fs_options_output)
                path_output_meta = base + '.yaml'
                # this fingerprint changes if input changes, or args, or kwargs
                fp = fingerprint(vaex.file.fingerprint(path_input, fs_options_input, fs=fs_input), args, kwargs)

                def write_fingerprint():
                    log.info('saving fingerprint %r for %s → %s conversion to %s', fp, path_input, path_output, path_output_meta)
                    with vaex.file.open(path_output_meta, 'w', fs_options=fs_options_output_, fs=fs_output) as f:
                        f.write(f"# this file exists so that we know when not to do the\n# {path_input} → {path_output} conversion\n")
                        vaex.utils.yaml_dump(f, {'fingerprint': fp})

                if not vaex.file.exists(path_output, fs_options=fs_options_output_, fs=fs_output):
                    log.info('file %s does not exist yet, running conversion %s → %s', path_output_meta, path_input, path_output)
                    value = callable(*args, **kwargs)
                    write_fingerprint()
                    return value

                if not vaex.file.exists(path_output_meta, fs_options=fs_options_output_, fs=fs_output):
                    log.info('file including fingerprint not found (%) or does not exist yet, running conversion %s → %s', path_output_meta, path_input, path_output)
                    value = callable(*args, **kwargs)
                    write_fingerprint()
                    return value

                # load fingerprint
                with vaex.file.open(path_output_meta, fs_options=fs_options_output_, fs=fs_output) as f:
                    output_meta = vaex.utils.yaml_load(f)

                if output_meta['fingerprint'] != fp:
                    log.info('fingerprint for %s is out of date, rerunning conversion to %s', path_input, path_output)
                    value = callable(*args, **kwargs)
                    write_fingerprint()
                    return value
                else:
                    log.info('fingerprint for %s did not change, reusing converted file %s', path_input, path_output)
            return call
        return wrapper2
    return wrapper1()



import vaex.dataset as dataset
import unittest
import vaex.expresso


class TestExpresso(unittest.TestCase):
	#def setUp(self):
	def test_ops(self):
		for binop in "+ - * / // % & | ^".split():
			vaex.expresso.validate_expression('a %s b' % binop, ['a', 'b'])
		for binop in "+ - * / // % & | ^".split():
			with self.assertRaises(NameError):
				vaex.expresso.validate_expression('a %s b' % binop, ['a'])
				vaex.expresso.validate_expression('a %s b' % binop, ['b'])
	def test_vars(self):
		with self.assertRaises(NameError):
			vaex.expresso.validate_expression('a+b', ['a'])
		vaex.expresso.validate_expression('a+b', ['a', 'b'])

	def test_functions(self):
		with self.assertRaises(NameError):
			vaex.expresso.validate_expression("f(a, b)", ['a'], ['f'])
		with self.assertRaises(NameError):
			vaex.expresso.validate_expression("f(a, b)", ['a', 'b'], ['g'])
		vaex.expresso.validate_expression("f(a, b)", ['a', 'b'], ['f'])

	def test_subsript(self):
		vaex.expresso.validate_expression("a[0]", ['a'])
		vaex.expresso.validate_expression("f(a[0])", ['a'], ['f'])
		vaex.expresso.validate_expression("a[0][0]", ['a'])
		with self.assertRaises(ValueError):
			vaex.expresso.validate_expression("a[0][b]", ['a', 'b'])
		with self.assertRaises(NameError):
			vaex.expresso.validate_expression("a[0]", [])



if __name__ == '__main__':
    unittest.main()



from __future__ import print_function
import os
import sys
import time


class ProgressBarBase(object):
    def __init__(self, min_value, max_value, format="%(percentage) 6.2f%% %(timeinfo)s cpu: %(cpu_usage)d%%"):
        self.min_value = min_value
        self.max_value = max_value
        self.format = format
        self.value = self.min_value
        self.fraction = 0
        self.prevfraction = 0

    def update_fraction(self):
        if (self.max_value-self.min_value) == 0:
            self.fraction = 1.
        else:
            self.fraction = (float(self.value)-self.min_value)/(self.max_value-self.min_value)

    def info(self):
        value = self.value
        if value == 0:
            self.prevtime = time.time()
            self.starttime = self.prevtime
            self.utime0, self.stime0 = os.times()[:2]
            self.walltime0 = time.time()
        currenttime = time.time()
        percentage = self.fraction * 100
        self.utime, self.stime = os.times()[:2]
        self.walltime  = time.time()
        cpu_time_delta = self.stime - self.stime0 + self.utime - self.utime0
        wall_time_delta = self.walltime - self.walltime0
        if wall_time_delta == 0:
            cpu_usage = 0  # it's formally infinity
        else:
            cpu_usage = cpu_time_delta / wall_time_delta * 100

        if (self.fraction > 0) and ((self.fraction-self.prevfraction) > 0):
            if self.fraction == 1:
                elapsedtime = (currenttime-self.starttime)
                seconds = elapsedtime
                minutes = seconds/60.
                hours = minutes/60.
                timeinfo = "elapsed time  : % 8.2fs = % 4.1fm = % 2.1fh" % (seconds, minutes, hours)
            else:
                #estimatedtime = (currenttime-self.starttime)/(self.fraction) * (1-self.fraction)
                #estimatedtime = (currenttime-self.prevtime)/(self.fraction-self.prevself.fraction) * (1-self.fraction)
                estimatedtime = (currenttime-self.prevtime) / (self.fraction) * (1-self.fraction)
                seconds = estimatedtime
                minutes = seconds/60.
                hours = minutes/60.
                timeinfo = "estimated time: % 8.2fs = % 4.1fm = % 2.1fh" % (seconds, minutes, hours)
        else:
            timeinfo = "estimated time: unknown                "
        return {"percentage":percentage, "timeinfo":timeinfo, "cpu_usage": cpu_usage}

    def __repr__(self):
        output = ''
        self.update_fraction()
        return self.format % self.info()

class ProgressBar(ProgressBarBase):
    """
        Implementation of progress bar on a console

        Usage::

            progressbar = ProgressBar(0, 100)
            progressbar.update(0)
            ....
            progressbar.update(100)

        By default, the progress bar writes to stderr, so it doesn't clutter up log files when piping stdout
    """
    def __init__(self, min_value, max_value, format="%(percentage) 6.2f%% %(timeinfo)s", width=40, barchar="#", emptychar="-", output=sys.stdout):
        """		
            :param min_value: minimum value for update(..)
            :param format: format specifier for the output
            :param width: width of the progress bar's (excluding extra text)
            :param barchar: character used to print the bar
            :param output: where to write the output to
        """
        super(ProgressBar, self).__init__(min_value, max_value, format=format)
        self.width = width
        self.barchar = barchar
        self.emptychar = emptychar
        self.output = output

    def update(self, value):
        self.value = value
        print(repr(self), file=self.output, end=' ')
        self.output.flush()

    def finish(self):
        if self.value != self.max_value:
            self.value = self.max_value
            print(repr(self), file=self.output, end=' ')
        self.output.flush()


    def __repr__(self):
        output = ''
        self.update_fraction()
        count = int(round(self.fraction * self.width))
        space = self.width - count
        bar = "[" + (self.barchar * count) + (self.emptychar * space) + "]"
        output = "\r" + bar + super(ProgressBar, self).__repr__()
        if self.fraction == 1:
            output += "\n" # last time print a newline char
        return output


class ProgressBarWidget(ProgressBarBase):
    def __init__(self, min_value, max_value, name=None):
        super(ProgressBarWidget, self).__init__(min_value, max_value)
        import ipywidgets as widgets
        from IPython.display import display
        self.bar = widgets.FloatProgress(min=self.min_value, max=self.max_value)
        self.text = widgets.Label(value='In progress...')
        self.widget = widgets.HBox([self.bar, self.text])
        # self.widget.description = repr(self)
        display(self.widget)

    def __call__(self, value):
        self.value = value
        self.bar.value = value
        self.text.value = repr(self)

    def update(self, value):
        self(value)

    def finish(self):
        self(self.max_value)



def docsubst(f):
    if f.__doc__:
        f.__doc__ = f.__doc__.format(**_doc_snippets)
    return f


_doc_snippets = {}
_doc_snippets["expression"] = "expression or list of expressions, e.g. df.x, 'x', or ['x, 'y']"
_doc_snippets["expression_one"] = "expression in the form of a string, e.g. 'x' or 'x+y' or vaex expression object, e.g. df.x or df.x+df.y "
_doc_snippets["expression_single"] = "if previous argument is not a list, this argument should be given"
_doc_snippets["binby"] = "List of expressions for constructing a binned grid"
_doc_snippets["limits"] = """description for the min and max values for the expressions, e.g. 'minmax' (default), '99.7%', [0, 10], or a list of, e.g. [[0, 10], [0, 20], 'minmax']"""
_doc_snippets["shape"] = """shape for the array where the statistic is calculated on, if only an integer is given, it is used for all dimensions, e.g. shape=128, shape=[128, 256]"""
_doc_snippets["percentile_limits"] = """description for the min and max values to use for the cumulative histogram, should currently only be 'minmax'"""
_doc_snippets["percentile_shape"] = """shape for the array where the cumulative histogram is calculated on, integer type"""
_doc_snippets["selection"] = """Name of selection to use (or True for the 'default'), or all the data (when selection is None or False), or a list of selections"""
_doc_snippets["selection1"] = """Name of selection to use (or True for the 'default'), or all the data (when selection is None or False)"""
_doc_snippets["delay"] = """Do not return the result, but a proxy for delayhronous calculations (currently only for internal use)"""
_doc_snippets["progress"] = """A callable that takes one argument (a floating point value between 0 and 1) indicating the progress, calculations are cancelled when this callable returns False"""
_doc_snippets["expression_limits"] = _doc_snippets["expression"]
_doc_snippets["grid"] = """If grid is given, instead if compuation a statistic given by what, use this Nd-numpy array instead, this is often useful when a custom computation/statistic is calculated, but you still want to use the plotting machinery."""
_doc_snippets["edges"] = """Currently for internal use only (it includes nan's and values outside the limits at borders, nan and 0, smaller than at 1, and larger at -1"""

_doc_snippets["healpix_expression"] = """Expression which maps to a healpix index, for the Gaia catalogue this is for instance 'source_id/34359738368', other catalogues may simply have a healpix column."""
_doc_snippets["healpix_max_level"] = """The healpix level associated to the healpix_expression, for Gaia this is 12"""
_doc_snippets["healpix_level"] = """The healpix level to use for the binning, this defines the size of the first dimension of the grid."""

_doc_snippets["return_stat_scalar"] = """Numpy array with the given shape, or a scalar when no binby argument is given, with the statistic"""
_doc_snippets["return_limits"] = """List in the form [[xmin, xmax], [ymin, ymax], .... ,[zmin, zmax]] or [xmin, xmax] when expression is not a list"""
_doc_snippets["cov_matrix"] = """List all convariance values as a double list of expressions, or "full" to guess all entries (which gives an error when values are not found), or "auto" to guess, but allow for missing values"""
_doc_snippets['propagate_uncertainties'] = """If true, will propagate errors for the new virtual columns, see :meth:`propagate_uncertainties` for details"""
_doc_snippets['note_copy'] = '.. note:: Note that no copy of the underlying data is made, only a view/reference is made.'
_doc_snippets['note_filter'] = '.. note:: Note that filtering will be ignored (since they may change), you may want to consider running :meth:`extract` first.'
_doc_snippets['inplace'] = 'Make modifications to self or return a new DataFrame'
_doc_snippets['return_shallow_copy'] = 'Returns a new DataFrame with a shallow copy/view of the underlying data'
_doc_snippets['chunk_size'] = 'Return an iterator with cuts of the object in lenght of this size'
_doc_snippets['chunk_size_export'] = 'Number of rows to be written to disk in a single iteration'
_doc_snippets['evaluate_parallel'] = 'Evaluate the (virtual) columns in parallel'
_doc_snippets['array_type'] = 'Type of output array, possible values are None (keep as is), "numpy" (ndarray), "xarray" for a xarray.DataArray, or "list"/"python" for a Python list'
_doc_snippets['ascii'] = 'Transform only ascii characters (usually faster).'
_doc_snippets['fs_options'] = 'see :func:`vaex.open` e.g. for S3 {"profile": "myproject"}'
_doc_snippets['fs'] = 'Pass a file system object directly, see :func:`vaex.open`'
